/*
Copyright (c) 2010, miya All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
* Neither the name of the copyright holders nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// PC用マンデルブロ集合のリアルタイムアニメーションプログラム
// リファレンスコード
// 2010/12/24 ver.1.00

#include <iostream>
#include <fstream>
#include <memory>
#include <sys/time.h>
#include <stdlib.h>
#include <pthread.h>
#include <X11/Xlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include <X11/Xutil.h>


// 画像の幅
#define XRES 480
// 画像の高さ
#define YRES 272
#define BPP 4
#define LOOP 360
#define WINDOW_NAME "Mandelbrot Set"


// 時間計測用関数。Tick値を返す（マイクロ秒）
static inline long long mcl_get_tick_count(void)
{
  struct timeval timeprof;
  gettimeofday(&timeprof, NULL);
  return ((long long)timeprof.tv_sec * (long long)1000000 + (long long)timeprof.tv_usec);
}


int main(int argc, char ** argv)
{
  unsigned int count, width, height;
  float xmin, xmax, ymin, ymax, x, y, a, b, a2, b2, dx, dy;
  unsigned int *pix;

  float zoom, cx, cy, cz, sx, sy, sz, speed;
  unsigned long long start_tick;
  unsigned int color_shift, cs_on;
  cx = 0.50485;
  cy = -0.6099;
  cz = 2.0;
  sx = 0.0;
  sy = 0.0;
  sz = 0.0;
  zoom = 2.0;
  speed = 0.97;
  color_shift = 0;
  cs_on = 0;

  Display * display;
  int screen;
  Window window;
  GC gc;
  XImage * image;
  XShmSegmentInfo shminfo;
  XSetWindowAttributes att;
  int depth;

  // Xサーバーに接続
  display = XOpenDisplay(NULL);

  // デフォルトスクリーンを取得
  screen = DefaultScreen(display);

  // スクリーンの深度を取得
  depth = DefaultDepth(display, screen);

  // ウィンドウを生成
  window = XCreateSimpleWindow(display, RootWindow(display, screen), 0, 0, XRES, YRES, 1, BlackPixel(display, screen), WhitePixel(display, screen));

  // ウィンドウマネージャをバイパス
  //att.override_redirect=True;
  //XChangeWindowAttributes(display, window, CWOverrideRedirect, &att);

  // ウィンドウ名をセット
  XStoreName(display, window, WINDOW_NAME);

  // 共有メモリー上のXImageを生成
  image = XShmCreateImage(display, DefaultVisual(display, screen), depth, ZPixmap, NULL, &shminfo, XRES, YRES);

  // フルカラーモードでなければ終了
  if (image->bitmap_pad != BPP * 8)
  {
    printf("Please execute in full-color mode.\n");
    exit(0);
  }

  // 共有メモリー・セグメントを生成
  shminfo.shmid = shmget(IPC_PRIVATE, image->bytes_per_line * image->height, IPC_CREAT|0777);

  // 共有メモリー・セグメントをアタッチ
  shminfo.shmaddr = image->data = (char*)shmat(shminfo.shmid, 0, 0);

  // Xサーバーからアクセス可能にする
  shminfo.readOnly = False;

  // 共有メモリーをアタッチすることをXサーバーに通知
  XShmAttach(display, &shminfo);

  // 出力バッファのsync
  XSync(display, False);

  // バッキングストアを有効にする
  att.backing_store = WhenMapped;
  XChangeWindowAttributes(display, window, CWBackingStore, &att);

  // GCを取得
  gc = DefaultGC(display, screen);

  // Copyする際にExposeイベントを送らない
  XSetGraphicsExposures(display, gc, False);

  // 描画色をセット
  XSetForeground(display, gc, BlackPixel(display, screen));

  // Exposeイベントを検出する
  XSelectInput(display, window, ExposureMask);

  // ウィンドウをマップ
  XMapWindow(display, window);

  // 出力バッファのflush
  XFlush(display);

  // 時間計測開始
  start_tick = mcl_get_tick_count();

  for (int k = 0; k < LOOP; k++)
  {
    zoom *= speed;
    color_shift++;

    xmin = cx - zoom * (float)XRES / (float)YRES;
    xmax = cx + zoom * (float)XRES / (float)YRES;
    ymin = cy - zoom;
    ymax = cy + zoom;

    width = XRES;
    height = YRES;

    dx = (xmax - xmin) / width;
    dy = (ymax - ymin) / height;

    y = ymin;

    for (unsigned int i = 0; i < height; i++)
    {
      pix = (unsigned int *)image->data;
      pix += XRES * i;
      x = xmin;

      for (unsigned int j = 0; j < width; j++)
      {
        a = x;
        b = y;
        a2 = a * a;
        b2 = b * b;
        count = 1024;
        while ((a2 + b2 <= 4) && count)
        {
          b = 2 * a * b - y;
          a = a2 - b2 - x;
          a2 = a * a;
          b2 = b * b;
          count--;
        }
        *pix = (count) ? ((count + color_shift) << 12) + count : ((a2 + b2) * 1024.0f);
        //*pix = (count << 12) + count;
        pix++;
        x += dx;
      }

      y += dy;
    }

    // Imageを書き込む
    XShmPutImage(display, window, gc, image, 0, 0, 0, 0, XRES, YRES, False);
    XFlush(display);
    XSync(display, False);

    // 演算終了を待つ
  }

  // 経過時間表示
  float calctime = (float)(mcl_get_tick_count() - start_tick) / 1000000.0f;
  float fps = (float)LOOP / calctime;
  printf("time: %f sec\n", calctime);
  printf("fps: %f\n", fps);

  // 共有メモリーをデタッチすることをXサーバーに通知
  XShmDetach(display, &shminfo);

  // XImageを解放
  XDestroyImage(image);

  // 共有メモリー・セグメントをデタッチ
  shmdt(shminfo.shmaddr);
  shmctl(shminfo.shmid, IPC_RMID, 0);

  // ウィンドウを解放
  XDestroyWindow(display, window);

  // Xサーバーから切断
  XCloseDisplay(display);

  return 0;
}
